from pydantic import BaseModel
from .nbschema import SchemaValidationResponse


class BaseError(BaseModel):
    pass


class InternalError(BaseError):
    details: str


class BadInput(BaseError):
    details: str


class PluginNotFound(BaseError):
    plugin_name: str


class NotebookNotFound(BaseError):
    notebook_id: str


class InputSchemaValidationError(BaseError):
    validation: SchemaValidationResponse



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jupyrest.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="jupyrest.proto",
    package="",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0ejupyrest.proto"4\n\x06Status\x12\x19\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x0b.StatusCode\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t"\x19\n\x08Notebook\x12\r\n\x05ipynb\x18\x01 \x01(\x0c"s\n\x18NotebookExecutionRequest\x12\x13\n\x0bnotebook_id\x18\x01 \x01(\t\x12\x13\n\x0bplugin_name\x18\x02 \x01(\t\x12\x12\n\nparameters\x18\x03 \x01(\x0c\x12\x19\n\x11parameterize_only\x18\x04 \x01(\x08"Q\n\x19NotebookExecutionResponse\x12\x17\n\x06status\x18\x01 \x01(\x0b\x32\x07.Status\x12\x1b\n\x08notebook\x18\x02 \x01(\x0b\x32\t.Notebook*\x88\x01\n\nStatusCode\x12\x06\n\x02OK\x10\x00\x12\x12\n\x0eINTERNAL_ERROR\x10\x01\x12\r\n\tBAD_INPUT\x10\x02\x12!\n\x1dINPUT_SCHEMA_VALIDATION_ERROR\x10\x03\x12\x16\n\x12NOTEBOOK_NOT_FOUND\x10\x04\x12\x14\n\x10PLUGIN_NOT_FOUND\x10\x05\x32Z\n\rWorkerService\x12I\n\x10\x65xecute_notebook\x12\x19.NotebookExecutionRequest\x1a\x1a.NotebookExecutionResponseb\x06proto3',
)

_STATUSCODE = _descriptor.EnumDescriptor(
    name="StatusCode",
    full_name="StatusCode",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="OK",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="INTERNAL_ERROR",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="BAD_INPUT",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="INPUT_SCHEMA_VALIDATION_ERROR",
            index=3,
            number=3,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="NOTEBOOK_NOT_FOUND",
            index=4,
            number=4,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="PLUGIN_NOT_FOUND",
            index=5,
            number=5,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=300,
    serialized_end=436,
)
_sym_db.RegisterEnumDescriptor(_STATUSCODE)

StatusCode = enum_type_wrapper.EnumTypeWrapper(_STATUSCODE)
OK = 0
INTERNAL_ERROR = 1
BAD_INPUT = 2
INPUT_SCHEMA_VALIDATION_ERROR = 3
NOTEBOOK_NOT_FOUND = 4
PLUGIN_NOT_FOUND = 5


_STATUS = _descriptor.Descriptor(
    name="Status",
    full_name="Status",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="code",
            full_name="Status.code",
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="details",
            full_name="Status.details",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=18,
    serialized_end=70,
)


_NOTEBOOK = _descriptor.Descriptor(
    name="Notebook",
    full_name="Notebook",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="ipynb",
            full_name="Notebook.ipynb",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=72,
    serialized_end=97,
)


_NOTEBOOKEXECUTIONREQUEST = _descriptor.Descriptor(
    name="NotebookExecutionRequest",
    full_name="NotebookExecutionRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="notebook_id",
            full_name="NotebookExecutionRequest.notebook_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="plugin_name",
            full_name="NotebookExecutionRequest.plugin_name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="parameters",
            full_name="NotebookExecutionRequest.parameters",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="parameterize_only",
            full_name="NotebookExecutionRequest.parameterize_only",
            index=3,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=99,
    serialized_end=214,
)


_NOTEBOOKEXECUTIONRESPONSE = _descriptor.Descriptor(
    name="NotebookExecutionResponse",
    full_name="NotebookExecutionResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="status",
            full_name="NotebookExecutionResponse.status",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="notebook",
            full_name="NotebookExecutionResponse.notebook",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=216,
    serialized_end=297,
)

_STATUS.fields_by_name["code"].enum_type = _STATUSCODE
_NOTEBOOKEXECUTIONRESPONSE.fields_by_name["status"].message_type = _STATUS
_NOTEBOOKEXECUTIONRESPONSE.fields_by_name["notebook"].message_type = _NOTEBOOK
DESCRIPTOR.message_types_by_name["Status"] = _STATUS
DESCRIPTOR.message_types_by_name["Notebook"] = _NOTEBOOK
DESCRIPTOR.message_types_by_name["NotebookExecutionRequest"] = _NOTEBOOKEXECUTIONREQUEST
DESCRIPTOR.message_types_by_name[
    "NotebookExecutionResponse"
] = _NOTEBOOKEXECUTIONRESPONSE
DESCRIPTOR.enum_types_by_name["StatusCode"] = _STATUSCODE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Status = _reflection.GeneratedProtocolMessageType(
    "Status",
    (_message.Message,),
    {
        "DESCRIPTOR": _STATUS,
        "__module__": "jupyrest_pb2"
        # @@protoc_insertion_point(class_scope:Status)
    },
)
_sym_db.RegisterMessage(Status)

Notebook = _reflection.GeneratedProtocolMessageType(
    "Notebook",
    (_message.Message,),
    {
        "DESCRIPTOR": _NOTEBOOK,
        "__module__": "jupyrest_pb2"
        # @@protoc_insertion_point(class_scope:Notebook)
    },
)
_sym_db.RegisterMessage(Notebook)

NotebookExecutionRequest = _reflection.GeneratedProtocolMessageType(
    "NotebookExecutionRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _NOTEBOOKEXECUTIONREQUEST,
        "__module__": "jupyrest_pb2"
        # @@protoc_insertion_point(class_scope:NotebookExecutionRequest)
    },
)
_sym_db.RegisterMessage(NotebookExecutionRequest)

NotebookExecutionResponse = _reflection.GeneratedProtocolMessageType(
    "NotebookExecutionResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _NOTEBOOKEXECUTIONRESPONSE,
        "__module__": "jupyrest_pb2"
        # @@protoc_insertion_point(class_scope:NotebookExecutionResponse)
    },
)
_sym_db.RegisterMessage(NotebookExecutionResponse)


_WORKERSERVICE = _descriptor.ServiceDescriptor(
    name="WorkerService",
    full_name="WorkerService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=438,
    serialized_end=528,
    methods=[
        _descriptor.MethodDescriptor(
            name="execute_notebook",
            full_name="WorkerService.execute_notebook",
            index=0,
            containing_service=None,
            input_type=_NOTEBOOKEXECUTIONREQUEST,
            output_type=_NOTEBOOKEXECUTIONRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_WORKERSERVICE)

DESCRIPTOR.services_by_name["WorkerService"] = _WORKERSERVICE

# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class StatusCode(_StatusCode, metaclass=_StatusCodeEnumTypeWrapper):
    pass
class _StatusCode:
    V = typing.NewType('V', builtins.int)
class _StatusCodeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_StatusCode.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    OK = StatusCode.V(0)
    INTERNAL_ERROR = StatusCode.V(1)
    BAD_INPUT = StatusCode.V(2)
    INPUT_SCHEMA_VALIDATION_ERROR = StatusCode.V(3)
    NOTEBOOK_NOT_FOUND = StatusCode.V(4)
    PLUGIN_NOT_FOUND = StatusCode.V(5)

OK = StatusCode.V(0)
INTERNAL_ERROR = StatusCode.V(1)
BAD_INPUT = StatusCode.V(2)
INPUT_SCHEMA_VALIDATION_ERROR = StatusCode.V(3)
NOTEBOOK_NOT_FOUND = StatusCode.V(4)
PLUGIN_NOT_FOUND = StatusCode.V(5)
global___StatusCode = StatusCode


class Status(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    CODE_FIELD_NUMBER: builtins.int
    DETAILS_FIELD_NUMBER: builtins.int
    code: global___StatusCode.V = ...
    details: typing.Text = ...
    def __init__(self,
        *,
        code : global___StatusCode.V = ...,
        details : typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"code",b"code",u"details",b"details"]) -> None: ...
global___Status = Status

class Notebook(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    IPYNB_FIELD_NUMBER: builtins.int
    ipynb: builtins.bytes = ...
    def __init__(self,
        *,
        ipynb : builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"ipynb",b"ipynb"]) -> None: ...
global___Notebook = Notebook

class NotebookExecutionRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NOTEBOOK_ID_FIELD_NUMBER: builtins.int
    PLUGIN_NAME_FIELD_NUMBER: builtins.int
    PARAMETERS_FIELD_NUMBER: builtins.int
    PARAMETERIZE_ONLY_FIELD_NUMBER: builtins.int
    notebook_id: typing.Text = ...
    plugin_name: typing.Text = ...
    parameters: builtins.bytes = ...
    parameterize_only: builtins.bool = ...
    def __init__(self,
        *,
        notebook_id : typing.Text = ...,
        plugin_name : typing.Text = ...,
        parameters : builtins.bytes = ...,
        parameterize_only : builtins.bool = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"notebook_id",b"notebook_id",u"parameterize_only",b"parameterize_only",u"parameters",b"parameters",u"plugin_name",b"plugin_name"]) -> None: ...
global___NotebookExecutionRequest = NotebookExecutionRequest

class NotebookExecutionResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    STATUS_FIELD_NUMBER: builtins.int
    NOTEBOOK_FIELD_NUMBER: builtins.int
    @property
    def status(self) -> global___Status: ...
    @property
    def notebook(self) -> global___Notebook: ...
    def __init__(self,
        *,
        status : typing.Optional[global___Status] = ...,
        notebook : typing.Optional[global___Notebook] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"notebook",b"notebook",u"status",b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"notebook",b"notebook",u"status",b"status"]) -> None: ...
global___NotebookExecutionResponse = NotebookExecutionResponse



import logging
from jupyrest.nbschema import SchemaValidationResponse
from typing import Any, Dict, Tuple, Iterable, Iterator, Optional, Type, Union
from nbformat.notebooknode import NotebookNode
from pydantic import BaseModel
from datetime import datetime
import json
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class BaseEvent(BaseModel):
    pass


class NotebookExecutionCreated(BaseEvent):
    notebook_id: str
    plugin_name: str
    parameters: Dict


class NotebookExecuted(BaseEvent):
    start_time: datetime
    end_time: datetime
    output_validation: SchemaValidationResponse
    exception: Optional[str] = None
    output_json_str: Optional[str] = None


class EventsDirectory:
    def __init__(self) -> None:
        self._id_to_event_type: Dict[str, Type[BaseEvent]] = {}
        self._event_type_to_id: Dict[Type[BaseEvent], str] = {}

    @classmethod
    def from_tuples(cls, event_types: Iterable[Tuple[str, Type[BaseEvent]]]):
        directory = cls()
        for event_id, event_type in event_types:
            directory.put(event_id=event_id, event_type=event_type)
        return directory

    def put(self, event_id: str, event_type: Type[BaseEvent]):
        if event_id in self._id_to_event_type:
            raise Exception(
                f"Event {self._id_to_event_type[event_id]} with id {event_id} already exists."
            )
        self._id_to_event_type[event_id] = event_type
        self._event_type_to_id[event_type] = event_id

    def get_event_type(self, event_id: str) -> Type[BaseEvent]:
        return self._id_to_event_type[event_id]

    def get_event_id(self, event_type: Type[BaseEvent]) -> str:
        return self._event_type_to_id[event_type]


class BaseEventStore(ABC):
    @abstractmethod
    def save_event(self, event: BaseEvent, key: Any):
        """Save an event to this event store using a key.

        Args:
            event (BaseEvent): event to save
        """
        pass

    @abstractmethod
    def get_event(self, key: Any) -> Optional[BaseEvent]:
        """Get a saved event under a key. If not exists, return None.

        Args:
            key (Any):

        Returns:
            Optional[BaseEvent]:
        """
        pass


class NotebookEventStore(BaseEventStore):

    directory = EventsDirectory.from_tuples(
        [
            ("notebook_execution_created", NotebookExecutionCreated),
            ("notebook_executed", NotebookExecuted),
        ]
    )

    event_id_key = "__event_id"

    def __init__(self, notebook: NotebookNode) -> None:
        self.notebook = notebook
        if "metadata" not in self.notebook:
            self.notebook["metadata"] = {}
        if "jupyrest" not in self.notebook["metadata"]:
            self.notebook["metadata"]["jupyrest"] = {}
        if "events" not in self.notebook["metadata"]["jupyrest"]:
            self.notebook["metadata"]["jupyrest"]["events"] = {}

    def save_event(self, event: BaseEvent, key=None):
        """Save an event in a notebook. Here the key is ignored and
        we use this class's EventsDirectory to generate a key for a given
        BaseEvent type.

        Args:
            event (BaseEvent):
            key ([type], optional): ignored
        """
        # convert event to json
        event_data = json.loads(event.json())
        # add event type info (if specified)
        event_id = self.directory.get_event_id(event_type=type(event))
        event_data[self.event_id_key] = event_id
        # save in notebook object
        self.notebook["metadata"]["jupyrest"]["events"][event_id] = event_data

    def get_event(self, key: Type[BaseEvent]) -> Optional[BaseEvent]:
        events = dict(self.notebook["metadata"]["jupyrest"]["events"])
        event_id = self.directory.get_event_id(event_type=key)
        event_json = events.get(event_id, None)
        if event_json is not None:
            return key.parse_obj(event_json)
        else:
            return None



from nbformat.notebooknode import NotebookNode
from .resolvers import BaseResolver
from .nbschema import NotebookSchemaProcessor, ConfigSections, SchemaValidationResponse
from .executors import BaseNotebookExeuctor
from typing import Optional, Dict, Type, Any
from abc import ABC, abstractmethod
import entrypoints
from papermill.parameterize import parameterize_notebook
from copy import deepcopy
import logging

logger = logging.getLogger(__name__)


class BasePlugin(ABC):
    """
    Contract for a Plugin.
    """

    @abstractmethod
    def parameterize_notebook(
        self, notebook_id: str, parameters: Any, notebook: NotebookNode
    ) -> NotebookNode:
        pass

    @abstractmethod
    def get_notebook_executor(self) -> BaseNotebookExeuctor:
        pass

    @abstractmethod
    def get_resolver(self) -> BaseResolver:
        pass

    @abstractmethod
    def get_nbschema(self) -> NotebookSchemaProcessor:
        pass

    @abstractmethod
    def validate_input(
        self, notebook_id: str, parameters: Any
    ) -> SchemaValidationResponse:
        pass

    @abstractmethod
    def validate_output(
        self, notebook_id: str, notebook: NotebookNode
    ) -> SchemaValidationResponse:
        pass


class PluginManager:

    DEFAULT_PLUGIN_NAME = "default"
    _plugins: Dict[str, BasePlugin] = {}

    def __init__(self) -> None:
        self.register_entry_points()

    def _get_plugin(self, plugin_name: str) -> Optional[BasePlugin]:
        return self._plugins.get(plugin_name, None)

    def register(self, plugin_name: str, plugin: BasePlugin):
        self._plugins[plugin_name] = plugin

    def register_entry_points(self):
        for entrypoint in entrypoints.get_group_all("jupyrest.plugins"):
            name = str(entrypoint.name)
            try:
                plugin: BasePlugin = entrypoint.load()
            except:
                logger.exception(f"Failed to load plugin {name}.")
            else:
                self.register(plugin_name=name, plugin=plugin)

    def load(self, plugin_name: str) -> Optional[BasePlugin]:
        plugin = self._get_plugin(plugin_name=plugin_name)
        if plugin is not None:
            return plugin
        else:
            return None


class JupyrestPlugin(BasePlugin):
    def __init__(
        self,
        resolver: BaseResolver,
        nbschema: NotebookSchemaProcessor,
        executor: BaseNotebookExeuctor,
    ) -> None:
        self._resolver = resolver
        self._nbschema = nbschema
        self._executor = executor

    def get_resolver(self) -> BaseResolver:
        return self._resolver

    def get_nbschema(self) -> NotebookSchemaProcessor:
        return self._nbschema

    def get_notebook_executor(self) -> BaseNotebookExeuctor:
        return self._executor

    def parameterize_notebook(
        self, notebook_id: str, parameters: Any, notebook: NotebookNode
    ) -> NotebookNode:
        config = self.get_resolver().resolve_config(notebook_id=notebook_id)
        parameters_copy = deepcopy(parameters)
        if config is not None and config.get(ConfigSections.INPUT, None) is not None:
            input_schema = config[ConfigSections.INPUT]
            parameters_copy = self.get_nbschema().inject_model_refs(
                input_schema, parameters_copy
            )
        # Parameterize notebooks with papermill, then fix it so it can work in jupyrest
        if "language" not in notebook.metadata.kernelspec:
            notebook.metadata.kernelspec[
                "language"
            ] = self.get_notebook_executor().get_kernelspec_language()
        # Papermill will do a deepcopy of the input notebook and
        # return the copy with the parameter cell
        new_notebook = parameterize_notebook(nb=notebook, parameters=parameters_copy)
        # Fix papermill's parameters metadata.
        # We reset the parameters to the original parameters prior to inject_model_refs
        new_notebook.metadata.papermill.parameters = parameters
        return new_notebook

    def validate_input(
        self, notebook_id: str, parameters: Any
    ) -> SchemaValidationResponse:
        notebook_config = self.get_resolver().resolve_config(notebook_id=notebook_id)
        return self.get_nbschema().validate_input(
            input=parameters, notebook_config=notebook_config
        )

    def validate_output(
        self, notebook_id: str, notebook: NotebookNode
    ) -> SchemaValidationResponse:
        output = self.get_nbschema().get_notebook_output(notebook=notebook)
        notebook_config = self.get_resolver().resolve_config(notebook_id=notebook_id)
        return self._nbschema.validate_output(
            output=output, notebook_config=notebook_config
        )



from pydantic import BaseModel, parse_obj_as
from typing import Callable, Type, Dict


class NamedModelConflict(Exception):
    pass


class NamedModel(BaseModel):
    """A NamedModel is a type that has a name. The name is
    stored in a `__ns__` class property. This information
    is a part of the objects JSON schema representation.

    Make a NamedModel:

    ```
    class MyNamedModel(NamedModel):
        __ns__ = "models/MyNamedModel"

        prop1: str
        prop2: int
    ```

    The benefit of a NamedModel is that it knows how to
    take a generic object and deserialize it to the correct
    subclass.

    ```
    import json

    # create the object
    model = MyNamedModel(prop1="foo", prop2=5)
    # convert to JSON string, then read it back as a dict
    model_json = json.loads(model.json())

    parsed_model = NamedModel.parse_obj(model_json)
    assert isinstance(parsed_model, MyNamedModel)
    ```
    """

    # a mapping from namespaces -> subclasses of NamedModel
    __nsmap__: Dict[str, Type["NamedModel"]] = {}
    # the name of this type (the namespace)
    __ns__: str

    @classmethod
    def _iter_subclasses(cls):
        """Iterate over the subclasses for a
        particular type `cls` including `cls` itself.
        """
        yield cls
        for sk in cls.__subclasses__():
            yield from sk._iter_subclasses()

    def dict(self, *args, **kwargs):
        """When we convert to a dict, we want
        to save the name of the model.
        """
        d = super().dict(*args, **kwargs)
        d[self.Config.NS_KEY] = getattr(self, self.Config.NS_KEY)
        return d

    @classmethod
    def parse_obj(cls, data, *args, **kwargs):
        NS_KEY = cls.Config.NS_KEY

        def data_has_ns_info(o):
            """This will tell us whether we should
            try parsing `data` or not
            """
            return isinstance(o, dict) and NS_KEY in data and isinstance(o[NS_KEY], str)

        def subclass_has_ns_info(sk):
            return hasattr(sk, NS_KEY) and isinstance(getattr(sk, NS_KEY), str)

        def subclass_has_name(name: str) -> Callable[[Type["NamedModel"]], bool]:
            def _check_name(sk: Type["NamedModel"]) -> bool:
                return getattr(sk, NS_KEY) == name

            return _check_name

        def get_subclass_for_ns(ns: str):
            if ns not in NamedModel.__nsmap__:
                # find the correct subclass
                sks = filter(subclass_has_ns_info, cls._iter_subclasses())
                sks = filter(subclass_has_name(data[NS_KEY]), sks)
                sks = set(sks)

                # we should only have at most 1 subclass
                # for a given name
                if len(sks) > 1:
                    raise NamedModelConflict(f"{sks} have conflicting names {ns}")
                elif len(sks) == 1:
                    sk = sks.pop()
                    NamedModel.__nsmap__[ns] = sk
                else:
                    return None

            return NamedModel.__nsmap__[ns]

        # if data has the namespace information we are looking for
        if data_has_ns_info(data):
            subclass = get_subclass_for_ns(data[NS_KEY])
            if subclass is not None:
                return parse_obj_as(subclass, data)

        return parse_obj_as(cls, data)

    class Config:

        NS_KEY: str = "__ns__"

        @staticmethod
        def schema_extra(schema, model: Type["NamedModel"]) -> None:
            """Augment the json schema to add the required
            namespace property.
            """
            NS_KEY = model.Config.NS_KEY

            # by default, the description is the class docstring
            # this is not relevant so we will remove it
            if "description" in schema and schema["description"] is not None:
                del schema["description"]

            if "properties" in schema and schema["properties"] is not None:
                schema["properties"][NS_KEY] = {"type": "string"}
            else:
                schema["required"] = {NS_KEY: {"type": "string"}}

            if "required" in schema and schema["required"] is not None:
                schema["required"].append(NS_KEY)
            else:
                schema["required"] = [NS_KEY]



from datetime import datetime
from jupyrest.nbschema import SchemaValidationResponse
from jupyrest.workers.generated.jupyrest_pb2 import Notebook
import logging
from logging import Logger
import json
from nbformat.notebooknode import NotebookNode
from jupyrest.domain import (
    NotebookEventStore,
    NotebookExecutionCreated,
    NotebookExecuted,
)
from typing import Any, Dict, Optional, Union, MutableMapping, Mapping
from uuid import uuid4
from jupyrest.errors import (
    BadInput,
    BaseError,
    PluginNotFound,
    InputSchemaValidationError,
    NotebookNotFound,
)
from jupyrest.plugin import PluginManager

logger = logging.getLogger(__name__)


class ExecuteNotebookLogAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        plugin_name = f'plugin_name={self.extra["plugin_name"]!r}'
        notebook_id = f'notebook_id={self.extra["notebook_id"]!r}'
        parameterize_only = f'parameterize_only={self.extra["parameterize_only"]!r}'
        context = "ExecuteNotebookInfo[{}, {}, {}]".format(
            plugin_name, notebook_id, parameterize_only
        )
        return f"{context} {msg}", kwargs


class Worker:
    """
    A Worker orchestrates the execution of notebooks and the retrieval of their outputs.
    """

    def __init__(self, plugin_man: PluginManager):
        self.plugin_man = plugin_man

    async def execute_notebook_async(
        self,
        plugin_name: str,
        notebook_id: str,
        parameters: Dict,
        parameterize_only: bool = False,
    ) -> Union[BaseError, NotebookNode]:
        # check args
        if not isinstance(plugin_name, str):
            return BadInput(details=f"invalid plugin_name parameter {plugin_name}")
        if not isinstance(notebook_id, str):
            return BadInput(details=f"invalid notebook_id parameter {notebook_id}")
        if not isinstance(parameterize_only, bool):
            return BadInput(
                details=f"parameterize_only is not of type bool, parameterize_only={parameterize_only}"
            )

        context = {
            "plugin_name": plugin_name,
            "notebook_id": notebook_id,
            "parameterize_only": parameterize_only,
        }
        log = ExecuteNotebookLogAdapter(logger, extra=context)

        # validate plugin exists
        plugin = self.plugin_man.load(plugin_name=plugin_name)
        if plugin is None:
            return PluginNotFound(plugin_name=plugin_name)
        log.info(f"Loaded plugin {plugin_name}")
        # validate notebook exists
        notebook = plugin.get_resolver().resolve_notebook(notebook_id=notebook_id)
        if notebook is None:
            return NotebookNotFound(notebook_id=notebook_id)

        # validate parameters and schema
        input_schema_validation = plugin.validate_input(
            notebook_id=notebook_id, parameters=parameters
        )
        if not input_schema_validation.is_valid:
            return InputSchemaValidationError(validation=input_schema_validation)

        notebook = plugin.parameterize_notebook(
            notebook_id=notebook_id, parameters=parameters, notebook=notebook
        )
        log.info(f"Parameterized notebook")
        event_store = NotebookEventStore(notebook=notebook)
        event_store.save_event(
            NotebookExecutionCreated(
                plugin_name=plugin_name, notebook_id=notebook_id, parameters=parameters,
            )
        )

        if parameterize_only:
            return event_store.notebook

        # execute notebook
        executor = plugin.get_notebook_executor()
        # TODO: remove start time and end time here if its never used
        start_time = datetime.utcnow()
        exception = await executor.execute_notebook_async(notebook=event_store.notebook)
        end_time = datetime.utcnow()
        log.info(
            f"Executor executed notebook. start_time={start_time.isoformat()}, end_time={end_time.isoformat()}"
        )
        output_json_str: Optional[str] = None
        output_validation = SchemaValidationResponse(is_valid=True)
        # check if notebook has output
        output_result = plugin.get_nbschema().get_notebook_output(
            notebook=event_store.notebook
        )
        if output_result.present:
            # get notebook output
            output_json_str = output_result.json_str
            # perform output validation
            notebook_config = plugin.get_resolver().resolve_config(
                notebook_id=notebook_id
            )
            output_validation = plugin.get_nbschema().validate_output(
                output=json.loads(output_json_str), notebook_config=notebook_config
            )
        # save execution event
        event_store.save_event(
            NotebookExecuted(
                start_time=start_time,
                end_time=end_time,
                exception=exception,
                output_validation=output_validation,
                output_json_str=output_json_str,
            )
        )
        return event_store.notebook



from json.decoder import JSONDecodeError

import logging
from uuid import uuid4
from typing import cast, Dict, Tuple
import pydantic
from typing import Optional, Any, Union
from nbformat.notebooknode import NotebookNode
from jupyrest.workers.base import Worker
from jupyrest.workers.generated import jupyrest_pb2 as grpc_models
from jupyrest.workers.generated import jupyrest_pb2_grpc as grpc_service
from jupyrest import errors
from jupyrest.nbschema import NbSchemaEncoder
from opentelemetry import trace
from opentelemetry.context import Context
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from grpc.aio import ServicerContext, Channel, Metadata
import json
import nbformat

logger = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)


class GrpcUtils:

    SPAN_METADATA_KEY = "otelspan"

    @classmethod
    def inject_span_context(cls, metadata: Metadata):
        span_context = {}
        TraceContextTextMapPropagator().inject(carrier=span_context)
        span_context_data = json.dumps(span_context)
        metadata[cls.SPAN_METADATA_KEY] = span_context_data

    @classmethod
    def extract_span_context(cls, context: ServicerContext) -> Optional[Context]:
        metadata = context.invocation_metadata()
        if metadata is None:
            return None
        if isinstance(metadata, tuple):
            metadata = Metadata.from_tuple(metadata)
        span_context_data = cast(str, metadata[cls.SPAN_METADATA_KEY])
        if span_context_data is not None:
            span_context_data = cast(Dict, json.loads(span_context_data))
            span_context = TraceContextTextMapPropagator().extract(
                carrier=span_context_data
            )
            return span_context

    @staticmethod
    def status_to_error(
        status: grpc_models.Status, notebook_id: str, plugin_name: str,
    ) -> Optional[errors.BaseError]:
        """
        Convert a grpc Status to a BaseError.

        :param status
        :type status: grpc_models.Status
        :param notebook_id
        :type notebook_id: models.NotebookIdentifier
        :param plugin_name
        :type plugin_name: str
        :return:
        :rtype: Optional[models.BaseError]
        """
        status_code = status.code
        details = status.details
        try:
            if status_code == grpc_models.StatusCode.OK:
                return None
            elif status_code == grpc_models.StatusCode.PLUGIN_NOT_FOUND:
                return errors.PluginNotFound(plugin_name=plugin_name)
            elif status_code == grpc_models.StatusCode.NOTEBOOK_NOT_FOUND:
                return errors.NotebookNotFound(notebook_id=notebook_id)
            elif status_code == grpc_models.StatusCode.BAD_INPUT:
                return errors.BadInput(details=status.details)
            elif status_code == grpc_models.StatusCode.INPUT_SCHEMA_VALIDATION_ERROR:
                validation = errors.SchemaValidationResponse(
                    is_valid=False, error=details
                )
                return errors.InputSchemaValidationError(validation=validation)
            elif status_code == grpc_models.StatusCode.INTERNAL_ERROR:
                return errors.InternalError(details=details)
        except pydantic.ValidationError as e:
            return errors.BadInput(details=str(e))

    @staticmethod
    def error_to_status(error: errors.BaseError) -> grpc_models.Status:
        """
        Convert a BaseError to the appripriate grpc Status.

        :param error:
        :type error: models.BaseError
        :return:
        :rtype: grpc_models.Status
        """
        if isinstance(error, errors.PluginNotFound):
            return grpc_models.Status(
                code=grpc_models.StatusCode.PLUGIN_NOT_FOUND, details=repr(error)
            )
        elif isinstance(error, errors.NotebookNotFound):
            return grpc_models.Status(
                code=grpc_models.StatusCode.NOTEBOOK_NOT_FOUND, details=repr(error),
            )
        elif isinstance(error, errors.InputSchemaValidationError):
            return grpc_models.Status(
                code=grpc_models.StatusCode.INPUT_SCHEMA_VALIDATION_ERROR,
                details=repr(error),
            )
        elif isinstance(error, errors.BadInput):
            return grpc_models.Status(
                code=grpc_models.StatusCode.BAD_INPUT, details=repr(error),
            )
        else:
            # unexcepted error will default to INTERNAL_ERROR
            return grpc_models.Status(
                code=grpc_models.StatusCode.INTERNAL_ERROR, details=repr(error)
            )


class GrpcWorkerServer(grpc_service.WorkerServiceServicer):
    """
    A GrpcServer that wraps a Worker object. Handles the convertion
    between grpc messages and jupyrest.models.
    """

    def __init__(self, worker: Worker) -> None:
        super().__init__()
        self._worker = worker

    async def execute_notebook(
        self, request: grpc_models.NotebookExecutionRequest, context: ServicerContext,
    ) -> grpc_models.NotebookExecutionResponse:
        span_context = GrpcUtils.extract_span_context(context)
        with tracer.start_as_current_span("grpc_request", context=span_context) as t:
            try:
                # parse grpc request
                plugin_name = str(request.plugin_name)
                notebook_id = str(request.notebook_id)

                try:
                    parameters = json.loads(request.parameters.decode("utf-8"))
                except JSONDecodeError as json_e:
                    logger.exception(
                        "An error occurred attempting to convert the parameters bytes into a json string."
                    )
                    return grpc_models.NotebookExecutionResponse(
                        status=grpc_models.Status(
                            code=grpc_models.StatusCode.BAD_INPUT, details=str(json_e),
                        ),
                        notebook=None,
                    )
                parameterize_only = request.parameterize_only

                # execute notebook
                response = await self._worker.execute_notebook_async(
                    plugin_name=plugin_name,
                    notebook_id=notebook_id,
                    parameters=parameters,
                    parameterize_only=parameterize_only,
                )

                if isinstance(response, errors.BaseError):
                    status = GrpcUtils.error_to_status(error=response)
                    return grpc_models.NotebookExecutionResponse(
                        status=status, notebook=None
                    )

                # convert notebook object to bytes
                notebook_data: bytes = nbformat.writes(
                    response, version=4, cls=NbSchemaEncoder,
                ).encode("utf-8")
                notebook = grpc_models.Notebook(ipynb=notebook_data)

                # return response
                status = grpc_models.Status(code=grpc_models.StatusCode.OK, details="")
                return grpc_models.NotebookExecutionResponse(
                    status=status, notebook=notebook
                )

            except Exception as e:
                logger.exception("An unhandled exception occured.")
                status = grpc_models.Status(
                    code=grpc_models.StatusCode.INTERNAL_ERROR, details=str(e),
                )
                return grpc_models.NotebookExecutionResponse(
                    status=status, notebook=None
                )


class GrpcWorkerClient:
    """
    An client adapter for a Worker.
    Wraps a grpc client to communicate with a running GrpcWorkerServer.
    """

    def __init__(
        self, channel: Channel, wait_for_ready: bool = False,
    ):
        super().__init__()
        self.stub = grpc_service.WorkerServiceStub(channel=channel)
        self.wait_for_ready = wait_for_ready

    async def execute_notebook_async(
        self,
        plugin_name: str,
        notebook_id: str,
        parameters: Any,
        parameterize_only=False,
    ) -> Union[errors.BaseError, NotebookNode]:

        with tracer.start_as_current_span("grpc_worker_client"):
            try:
                try:
                    parameter_bytes = json.dumps(parameters).encode("utf-8")
                except TypeError as t:
                    # catch errors when attempting to convert parameters to json string
                    return errors.BadInput(details=str(t))

                grpc_request = grpc_models.NotebookExecutionRequest(
                    notebook_id=notebook_id,
                    plugin_name=plugin_name,
                    parameters=parameter_bytes,
                    parameterize_only=parameterize_only,
                )
                # create a Metadata object and insert
                # the current span context in it
                metadata = Metadata()
                GrpcUtils.inject_span_context(metadata)
                grpc_response = await self.stub.execute_notebook(
                    grpc_request, wait_for_ready=self.wait_for_ready, metadata=metadata
                )
                grpc_response = cast(
                    grpc_models.NotebookExecutionResponse, grpc_response
                )
                if grpc_response.status.code != grpc_models.StatusCode.OK:
                    # return appropriate error is StatusCode != OK
                    error = GrpcUtils.status_to_error(
                        status=grpc_response.status,
                        notebook_id=notebook_id,
                        plugin_name=plugin_name,
                    )
                    if error is not None:
                        return error

                notebook: NotebookNode = nbformat.reads(
                    grpc_response.notebook.ipynb.decode("utf-8"),
                    as_version=nbformat.NO_CONVERT,
                )
                return notebook
            except Exception as e:
                logger.exception("An unhandled exception occurred.")
                return errors.InternalError(details=str(e))



from jupyrest.workers.base import Worker
from jupyrest.executors import IPythonNotebookExecutor
from jupyrest.nbschema import NotebookSchemaProcessor
from jupyrest.resolvers import LocalDirectoryResolver
import typer
from enum import Enum
from typing import Optional
from pathlib import Path
from jupyrest.plugin import PluginManager, JupyrestPlugin
from jupyrest.workers.grpc import GrpcWorkerServer
from jupyrest.workers.http import create_dev_app
from jupyrest.workers.generated import jupyrest_pb2_grpc
import grpc
import asyncio
import uvicorn
import logging
from opentelemetry import trace


class WorkerType(str, Enum):
    grpc = "grpc"
    http = "http"


HTTP_PORT = 5050
GRPC_PORT = 50051
MAX_MESSAGE_LENGTH = 100 * 1024 * 1024  # 100 MB

cli_app = typer.Typer(name="jupyrest cli")
worker_app = typer.Typer()

cli_app.add_typer(worker_app, name="worker")


async def serve_grpc(worker: Worker, port: int):

    server = grpc.aio.server(
        options=[
            ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
            ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH),
        ]
    )
    jupyrest_pb2_grpc.add_WorkerServiceServicer_to_server(
        GrpcWorkerServer(worker), server
    )
    listen_addr = f"[::]:{port}"
    tracer = trace.get_tracer(__name__)
    with tracer.start_as_current_span("start grpc server"):
        server.add_insecure_port(listen_addr)
        logging.info(f"Starting grpc worker server on {listen_addr}")
        await server.start()
        try:
            await server.wait_for_termination()
        except KeyboardInterrupt:
            # Shuts down the server with 0 seconds of grace period. During the
            # grace period, the server won't accept new connections and allow
            # existing RPCs to continue within the grace period.
            await server.stop(0)


@worker_app.command("start")
def start(
    worker_type: WorkerType,
    notebooks_dir: Optional[Path] = typer.Option(None),
    port: Optional[int] = typer.Option(None, help="Port to serve worker on."),
    fluentd: bool = False,
):

    plugin_man = PluginManager()
    if notebooks_dir is not None:
        plugin_man.register(
            plugin_name=PluginManager.DEFAULT_PLUGIN_NAME,
            plugin=JupyrestPlugin(
                resolver=LocalDirectoryResolver(notebooks_dir=notebooks_dir),
                nbschema=NotebookSchemaProcessor(),
                executor=IPythonNotebookExecutor(),
            ),
        )
    worker = Worker(plugin_man=plugin_man)

    if worker_type == WorkerType.grpc:
        if port is None:
            port = GRPC_PORT
        loop = asyncio.get_event_loop()
        coro = serve_grpc(worker=worker, port=port)
        loop.run_until_complete(coro)

    elif worker_type == WorkerType.http:
        if port is None:
            port = HTTP_PORT
        http_app = create_dev_app(worker=worker)
        typer.echo(f"Starting http worker server on {port}")
        uvicorn.run(app=http_app, port=port)  # type: ignore



from abc import abstractmethod, ABC
import nbformat
from typing import Optional, Dict, Any
from pathlib import Path
import json

from nbformat.notebooknode import NotebookNode
from .nbschema import NotebookConfig
from papermill import __version__
from opentelemetry import trace
from azure.storage.blob import ContainerClient

tracer = trace.get_tracer(__name__)


def load_notebook_node(nb: NotebookNode) -> NotebookNode:
    """Returns a notebook object with papermill and cell tag metadata.
    This function was taken from papermill.iorw.load_notebook_node.

    Args:
        nb (NotebookNode)

    Returns:
        NotebookNode
    """

    if not hasattr(nb.metadata, "papermill"):
        nb.metadata["papermill"] = {
            "default_parameters": dict(),
            "parameters": dict(),
            "environment_variables": dict(),
            "version": __version__,
        }

    for cell in nb.cells:
        if not hasattr(cell.metadata, "tags"):
            cell.metadata["tags"] = []  # Create tags attr if one doesn't exist.

        if not hasattr(cell.metadata, "papermill"):
            cell.metadata["papermill"] = dict()

    return nb


class BaseResolver(ABC):
    @abstractmethod
    def resolve_notebook(self, notebook_id: str) -> Optional[nbformat.NotebookNode]:
        pass

    @abstractmethod
    def resolve_config(self, notebook_id: str) -> Optional[NotebookConfig]:
        pass


class LocalDirectoryResolver(BaseResolver):

    _config_cache: Optional[Dict[Path, Dict]] = None

    def __init__(self, notebooks_dir: Path):
        if not notebooks_dir.exists() or not notebooks_dir.is_dir():
            raise ValueError(f"{notebooks_dir} needs to be a valid directory")
        self._notebooks_dir = notebooks_dir

    @classmethod
    def _read_file(cls, path: Path) -> str:
        return path.read_bytes().decode("utf-8-sig")

    @classmethod
    def _read_config(cls, path: Path) -> Optional[NotebookConfig]:
        if not path.exists():
            return None
        if cls._config_cache is None:
            cls._config_cache = dict()
        if path not in cls._config_cache:
            text = cls._read_file(path=path)
            config_dict: Dict[str, Any] = json.loads(text)
            cls._config_cache[path] = config_dict
        return cls._config_cache[path]

    def get_notebook_file(self, notebook_id: str) -> Path:
        return self.notebooks_dir / f"{notebook_id}.ipynb"

    def get_config_file(self, notebook_id: str) -> Path:
        notebook_file = self.get_notebook_file(notebook_id=notebook_id)
        return notebook_file.parent / f"{notebook_file.stem}.config.json"

    @property
    def notebooks_dir(self) -> Path:
        return self._notebooks_dir

    def resolve_notebook(self, notebook_id: str) -> Optional[nbformat.NotebookNode]:
        notebook_file = self.get_notebook_file(notebook_id=notebook_id)
        if notebook_file.exists() and notebook_file.is_file():
            nb = nbformat.reads(
                notebook_file.read_text(), as_version=nbformat.NO_CONVERT
            )
            return load_notebook_node(nb=nb)

    def resolve_config(self, notebook_id: str) -> Optional[NotebookConfig]:
        config_file = self.get_config_file(notebook_id=notebook_id)
        return self._read_config(path=config_file)


class BlobNotFoundException(Exception):
    pass


class AzureStorageResolver(BaseResolver):
    def __init__(self, container_client: ContainerClient, prefix: Optional[str] = None):
        self.notebook_container_client = container_client
        self.prefix = prefix or ""

    def make_notebook_uri(self, notebook_id: str):
        return f"{self.prefix}{notebook_id}.ipynb"

    def make_config_uri(self, notebook_id: str):
        return f"{self.prefix}{notebook_id}.config.json"

    def get_blob_contents(self, uri: str) -> str:
        blob_client = self.notebook_container_client.get_blob_client(uri)

        if not blob_client.exists():
            raise BlobNotFoundException(f"Blob with uri {uri} does not exist")

        download_blob = blob_client.download_blob()
        blob_contents = download_blob.readall()
        if isinstance(blob_contents, bytes):
            blob_contents = blob_contents.decode()
        return blob_contents

    def resolve_notebook(self, notebook_id: str) -> Optional[nbformat.NotebookNode]:
        with tracer.start_as_current_span("azstorage.resolve_notebook"):
            notebook_obj: Optional[nbformat.NotebookNode] = None
            notebook_file_uri = f"{notebook_id}.ipynb"
            try:
                notebook_contents = self.get_blob_contents(notebook_file_uri)
                notebook_obj = nbformat.reads(
                    s=notebook_contents, as_version=nbformat.NO_CONVERT
                )
            except BlobNotFoundException:
                pass

            if notebook_obj is not None:
                notebook_obj = load_notebook_node(notebook_obj)
            return notebook_obj

    def resolve_config(self, notebook_id: str) -> Optional[NotebookConfig]:
        with tracer.start_as_current_span("azstorage.resolve_config"):
            config_dict: Optional[Dict[str, Any]] = None
            config_file_uri = f"{notebook_id}.config.json"
            try:
                config_contents = self.get_blob_contents(config_file_uri)
                config_dict = json.loads(config_contents)
            except BlobNotFoundException:
                pass

            return config_dict



import json
from nbformat.notebooknode import NotebookNode
import scrapbook as sb
from jsonschema import Draft7Validator, RefResolver
from jsonschema.exceptions import best_match
from typing import Optional, Dict, Type, Union, Any, Set
from urllib.parse import urlparse
from pydantic import BaseModel
from datetime import datetime, date
from papermill.translators import papermill_translators, PythonTranslator
from enum import Enum


class ConfigSections(str, Enum):
    INPUT = "input"
    OUTPUT = "output"


NotebookConfig = Dict


class SchemaValidationResponse(BaseModel):
    is_valid: bool
    error: Optional[str] = None


class NbSchemaBase(BaseModel):
    """
    Base class for all NbSchema models.
    Inherits from Pydantic BaseModel.
    """

    pass


class ModelCollection:
    """
    Stores NbSchemaBase implementations along with their aliases.
    """

    def __init__(self) -> None:
        self._map: Dict[str, Type[NbSchemaBase]] = {}

    def add_model(self, alias: str, model_type: Type[NbSchemaBase], overwrite=False):
        """
        Add a model class and its alias to the collection.
        Duplicate aliases will be a no-op unless overwrite is True.
        """
        if alias not in self._map or overwrite:
            self._map[alias] = model_type

    def has_alias(self, alias: str) -> bool:
        return alias in self._map

    def get_model(self, alias: str) -> Type[NbSchemaBase]:
        if not self.has_alias(alias=alias):
            raise KeyError(f"Model for alias {alias} not found.")
        return self._map[alias]

    def is_empty(self) -> bool:
        return len(self._map.keys()) == 0


class NbSchemaEncoder(json.JSONEncoder):
    """
    JSONEncoder that can encode NbSchemaBase objects to json().
    This also encodes datetime objects to strings in iso format.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, NbSchemaBase):
            return json.loads(o.json())
        elif isinstance(o, (datetime, date)):
            return o.isoformat()
        else:
            return json.JSONEncoder.default(self, o)


class NbSchemaTranslator(PythonTranslator):
    """
    Translator for NbSchemaBase objects.
    """

    @classmethod
    def translate(cls, val):
        if isinstance(val, NbSchemaBase):
            return f"{val!r}"
        return super().translate(val)


class OutputResult(BaseModel):
    present: bool
    json_str: str


class NotebookSchemaProcessor:

    OUTPUTS_KEY = "nbschema_outputs"
    SCHEME = "nbschema"

    def __init__(self, models: Optional[ModelCollection] = None) -> None:
        self._models = models or ModelCollection()
        self._ref_resolver = RefResolver(
            "", {}, handlers={self.SCHEME: self._resolve_ref}
        )
        papermill_translators.register("python", NbSchemaTranslator)

    @classmethod
    def _uri_to_alias(cls, uri: str):
        scheme = urlparse(uri).scheme
        if scheme != cls.SCHEME:
            raise ValueError(f"Invalid uri scheme {scheme}")
        return uri.replace(f"{cls.SCHEME}://", "")

    def _get_validator(self, schema: Optional[Dict] = None):
        Draft7Validator.check_schema(schema)
        if not self._models.is_empty():
            return Draft7Validator(schema=schema, resolver=self._ref_resolver)
        else:
            return Draft7Validator(schema=schema)

    def _validate_instance(
        self, instance: Optional[Dict], schema: Optional[Dict] = None
    ) -> SchemaValidationResponse:
        # if we have a schema to validate against
        if schema is not None:
            # Check that this is a valid JSONSchema. Raise an exception if it is not.
            validator = self._get_validator(schema=schema)
            error = best_match(validator.iter_errors(instance=instance))
            if error is not None:
                return SchemaValidationResponse(is_valid=False, error=str(error))
        return SchemaValidationResponse(is_valid=True)

    def _resolve_ref(self, ref: str):
        alias = self._uri_to_alias(uri=ref)
        return self._models.get_model(alias=alias).schema()

    @property
    def models(self):
        return self._models

    def inject_model_refs(self, schema: Dict, payload: Dict):
        """
        This function assumes that `payload` adheres to the JSON Schema
        defined by `schema`.

        Does not support "anyOf", "allOf", "oneOf" or "contains".

        Recursively walk the key value pairs of each of the input
        dicts (schema and payload) in lockstep.

        Wherever `schema` has a $ref to an nbschema model, replace
        the corresponding object in `payload` with a instance of the
        NbSchemaBase class that is referenced. NbSchemaBase objects know
        how to create themselves from a dict.

        Returns a payload object with NbSchemaBase objects injected if any.
        """
        if schema is None or payload is None:
            return payload
        # Case 1: schema is a $ref to nbschema://
        ref = schema.get("$ref", None)
        if ref is not None:
            u = urlparse(ref)
            if u.scheme == "nbschema":
                alias = self._uri_to_alias(uri=ref)
                klass = self._models.get_model(alias=alias)
                obj = klass.parse_obj(payload)
                return obj

        # Case 2: schema "type" is an object or array.
        schema_type = schema.get("type", None)
        if schema_type is not None:
            if schema_type == "object":
                # if object, recursive call on kv pairs
                props = schema.get("properties", {})
                for prop in props:
                    if prop in payload:
                        new_schema = props[prop]
                        new_payload = payload[prop]
                        payload[prop] = self.inject_model_refs(new_schema, new_payload)
                return payload
            elif schema_type == "array":
                # if list, recursive call on list items
                item_schema = schema.get("items", None)
                if item_schema is not None:
                    objs = []
                    for item in payload:
                        obj = self.inject_model_refs(item_schema, item)
                        objs.append(obj)
                    return objs
        return payload

    def validate_input(
        self, input: Dict, notebook_config: Optional[NotebookConfig]
    ) -> SchemaValidationResponse:
        input_schema = (
            notebook_config.get(ConfigSections.INPUT, None) if notebook_config else None
        )
        return self._validate_instance(instance=input, schema=input_schema,)

    def validate_output(
        self, output, notebook_config: Optional[NotebookConfig]
    ) -> SchemaValidationResponse:
        output_schema = (
            notebook_config.get(ConfigSections.OUTPUT, None)
            if notebook_config
            else None
        )
        return self._validate_instance(instance=output, schema=output_schema)

    @classmethod
    def save_output(cls, data: Union[str, NbSchemaBase], **kwargs):
        """
        Simple wrapper around scrapbook's glue() function.
        The only difference is that the `name` parameter
        is set to OUTPUTS_KEY. The data to save should
        be a string compatible with the "application/json"
        mimetype or it must be an NbSchemaBase.

        Usage (in a notebook):
        >>> save_output(data='{"json_key": "json_value"}')
        >>> save_output(data=json.dumps(my_obj))

        :param data: data to save
        :param kwagrs: passed on to scrapbook.glue
        """
        json_str = None
        if isinstance(data, str):
            json_str = data
        else:
            json_str = json.dumps(data, cls=NbSchemaEncoder)
        # test that this string is valid JSON
        json.loads(json_str)
        sb.glue(cls.OUTPUTS_KEY, json_str, **kwargs)

    def get_notebook_output(self, notebook: NotebookNode) -> OutputResult:
        """
        Get outputs from an executed notebook.
        """
        nb = sb.read_notebook(notebook)
        if self.OUTPUTS_KEY in nb.scraps.data_dict:
            json_str = nb.scraps.data_dict[self.OUTPUTS_KEY]
            return OutputResult(present=True, json_str=json_str)
        else:
            return OutputResult(present=False, json_str="")



from abc import ABC, abstractmethod
from datetime import datetime
from dataclasses import dataclass
from nbformat.notebooknode import NotebookNode
from typing import Optional
from nbclient.client import NotebookClient
from nbclient.exceptions import CellExecutionError, CellTimeoutError
import logging
from opentelemetry import trace

logger = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)


class BaseNotebookExeuctor(ABC):
    @abstractmethod
    def get_kernelspec_language(self) -> str:
        pass

    @abstractmethod
    async def execute_notebook_async(self, notebook: NotebookNode) -> Optional[str]:
        """Executes a notebook in place. Returns an exception string if any.

        Args:
            notebook (NotebookNode): notebook to execute

        Returns:
            Optional[str]: exception
        """
        pass


class OtelNotebookClient(NotebookClient):
    """A NotebookClient that emits OpenTelemetry Spans."""

    async def async_execute(self, reset_kc: bool = False, **kwargs) -> NotebookNode:
        with tracer.start_as_current_span("nbclient.async_execute"):
            return await super().async_execute(reset_kc=reset_kc, **kwargs)


class IPythonNotebookExecutor(BaseNotebookExeuctor):
    def __init__(
        self, kernel_name="python3", timeout_seconds=600, language="python"
    ) -> None:
        self._kernel_name = kernel_name
        self._timeout_seconds = timeout_seconds
        self._language = language

    def get_kernelspec_language(self) -> str:
        return self._language

    async def execute_notebook_async(self, notebook: NotebookNode) -> Optional[str]:
        exception: Optional[str] = None
        try:
            await OtelNotebookClient(
                nb=notebook,
                timeout=self._timeout_seconds,
                kernel_name=self._kernel_name,
                log=logger,
            ).async_execute()
        except CellExecutionError as cee:
            # handle cases where the notebook calls sys.exit(0),
            # which is considered successful.
            is_sys_exit_0 = cee.ename == "SystemExit" and (
                cee.evalue == "" or cee.evalue == "0"
            )
            if not is_sys_exit_0:
                exception = str(cee)
        except CellTimeoutError as cte:
            exception = str(cte)
        return exception



from jupyrest.domain import (
    NotebookEventStore,
    NotebookExecuted,
    NotebookExecutionCreated,
)
from nbformat import NotebookNode
import nbformat
import json
from starlette.responses import HTMLResponse
from jupyrest.plugin import PluginManager
from jupyrest.workers.base import Worker
from pydantic import BaseModel
from jupyrest.nbschema import NbSchemaEncoder
from jupyrest.errors import (
    BadInput,
    BaseError,
    InternalError,
    PluginNotFound,
    NotebookNotFound,
    InputSchemaValidationError,
)
from typing import Dict
from fastapi import FastAPI, HTTPException
from uuid import uuid4
from nbconvert import HTMLExporter
from opentelemetry import trace

tracer = trace.get_tracer(__name__)


class NotebookRequest(BaseModel):
    notebook: str
    parameters: Dict


class NotebookResponse(BaseModel):
    id: str
    status: str
    notebook: str
    parameters: Dict


def notebook_to_html(notebook: NotebookNode, report_mode=False) -> str:
    with tracer.start_as_current_span("notebook_to_html"):
        exporter = HTMLExporter()
        if report_mode:
            exporter = HTMLExporter(exclude_output_prompt=True, exclude_input=True)
        (body, _) = exporter.from_notebook_node(notebook)
        return body


def notebook_to_str(notebook: NotebookNode) -> str:
    with tracer.start_as_current_span("notebook_to_str"):
        return nbformat.writes(
            notebook, version=nbformat.NO_CONVERT, cls=NbSchemaEncoder
        )


def create_dev_app(worker: Worker):
    app = FastAPI(title="Jupyrest Web Server", debug=True)
    notebooks: Dict[str, NotebookNode] = {}

    plugin_name = PluginManager.DEFAULT_PLUGIN_NAME

    def to_http_exception(error: BaseError) -> HTTPException:
        if isinstance(error, PluginNotFound):
            return HTTPException(
                status_code=404, detail=f"Plugin {error.plugin_name} not found."
            )
        elif isinstance(error, NotebookNotFound):
            return HTTPException(
                status_code=404, detail=f"Notebook {error.notebook_id} not found."
            )
        elif isinstance(error, InputSchemaValidationError):
            return HTTPException(
                status_code=400,
                detail=f"Input schema validation error: {error.validation.error}",
            )
        elif isinstance(error, BadInput):
            return HTTPException(status_code=400, detail=f"Bad input: {error.details}",)
        elif isinstance(error, InternalError):
            return HTTPException(status_code=500, detail=error.details)
        else:
            return HTTPException(status_code=500)

    @app.post("/api/NotebookExecutions")
    async def post(req: NotebookRequest) -> NotebookResponse:
        # create NotebookExecutionRequest
        notebook_id = req.notebook
        parameters = req.parameters
        execution_id = str(uuid4())
        # execute notebook
        result = await worker.execute_notebook_async(
            plugin_name=plugin_name,
            notebook_id=notebook_id,
            parameters=parameters,
            parameterize_only=False,
        )

        if isinstance(result, BaseError):
            raise to_http_exception(result)

        # save execution
        notebooks[execution_id] = result

        event_store = NotebookEventStore(notebook=result)
        notebook_executed = event_store.get_event(NotebookExecuted)
        assert isinstance(notebook_executed, NotebookExecuted)
        status = "COMPLETED" if notebook_executed.exception is None else "FAILED"
        response = NotebookResponse(
            id=execution_id,
            status=status,
            notebook=req.notebook,
            parameters=req.parameters,
        )
        return response

    @app.get("/api/NotebookExecutions")
    async def get(
        executionId: str,
        html: bool = False,
        report: bool = False,
        output: bool = False,
        ipynb: bool = False,
        view_html: bool = False,
    ):

        if executionId not in notebooks:
            raise HTTPException(
                status_code=404, detail=f"Execution {executionId} not found."
            )
        notebook = notebooks[executionId]
        event_store = NotebookEventStore(notebook=notebook)
        execution_created = event_store.get_event(NotebookExecutionCreated)
        notebook_executed = event_store.get_event(NotebookExecuted)
        assert isinstance(notebook_executed, NotebookExecuted)
        assert isinstance(execution_created, NotebookExecutionCreated)
        status = "COMPLETED" if notebook_executed.exception is None else "FAILED"
        response = NotebookResponse(
            id=executionId,
            status=status,
            notebook=execution_created.notebook_id,
            parameters=execution_created.parameters,
        ).dict()
        response["exception"] = notebook_executed.exception
        if view_html:
            return HTMLResponse(notebook_to_html(notebook=notebook, report_mode=report))
        if html:
            response["html"] = notebook_to_html(notebook=notebook, report_mode=report)
        if output:
            if notebook_executed.output_json_str is None:
                response["output"] = None
            else:
                response["output"] = json.loads(notebook_executed.output_json_str)
        if ipynb:
            response["ipynb"] = notebook_to_str(notebook=notebook)
        return response

    return app



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jupyrest', 'jupyrest.workers', 'jupyrest.workers.generated']

package_data = \
{'': ['*']}

install_requires = \
['azure-storage-blob>=12.8.1,<13.0.0',
 'black==21.7b0',
 'click>=7.1.2,<8.0.0',
 'cryptography<=3.4.7',
 'entrypoints>=0.3,<0.4',
 'fastapi>=0.65.2,<0.66.0',
 'grpcio-tools>=1.38.0,<2.0.0',
 'grpcio>=1.38.0,<2.0.0',
 'ipykernel>=6.0.1,<7.0.0',
 'ipython>=7.25.0,<8.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'jupyter-client<=6.1.12',
 'nbclient>=0.5.1,<0.6.0',
 'nbconvert==5.6.1',
 'nbformat>=5.1.3,<6.0.0',
 'opentelemetry-api>=1.5.0,<2.0.0',
 'papermill>=2.2,<2.3',
 'pydantic>=1.8,<2.0',
 'scrapbook>=0.5.0,<0.6.0',
 'setuptools-rust>=0.12.1,<0.13.0',
 'textwrap3>=0.9.2,<0.10.0',
 'typer>=0.3.2,<0.4.0',
 'uvicorn>=0.14.0,<0.15.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.7.4,<4.0.0'],
 ':sys_platform == "win32"': ['pywin32==301']}

setup_kwargs = {
    'name': 'jupyrest',
    'version': '0.1.0',
    'description': 'A tool to expose Jupyter notebooks as a REST API.',
    'long_description': None,
    'author': 'Koushik Krishnan',
    'author_email': 'kokrishn@microsoft.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



import warnings

warnings.simplefilter("ignore", FutureWarning)
warnings.simplefilter("ignore", DeprecationWarning)

from .nbschema import NotebookSchemaProcessor

save_output = NotebookSchemaProcessor.save_output
